"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
from servalcat.utils import logger
from servalcat.utils import model
from servalcat.utils import hkl
from servalcat.utils import restraints
import os
import shutil
import glob
import re
import subprocess
import gemmi
import numpy
import numpy.lib.recfunctions
import gzip

def splitext(path):
    if path.endswith((".bz2",".gz")):
        return os.path.splitext(path[:path.rindex(".")])
    else:
        return os.path.splitext(path)
# splitext()

def rotate_file(filename, copy=False):
    if not os.path.exists(filename): return

    # make list [ [filename, number], ... ]
    old_list = []
    dot_files = glob.glob(filename + ".*")
    for f in dot_files:
        suffix = f.replace(filename+".", "")
        try:
            i = int(suffix)
            if str(i) == suffix: # ignore if suffix was such as 003...
                old_list.append([f, i])
        except ValueError as e:
            continue

    old_list.sort(key=lambda x: x[1])

    # rotate files
    for f, i in reversed(old_list):
        logger.writeln("Rotating file: {}".format(f))
        os.rename(f, "%s.%d" % (f[:f.rfind(".")], i+1))

    if copy:
        shutil.copyfile(filename, filename + ".1")
    else:
        os.rename(filename, filename + ".1")

    return filename + ".1"
# rotate_file()

def check_model_format(xyzin):
    # TODO check format actually
    # TODO mmjson is possible?
    ext = splitext(xyzin)[1]
    if ext.endswith("cif"):
        return ".mmcif"
    else:
        return ".pdb"
# check_model_format()

def write_mmcif(st, cif_out, cif_ref=None):
    """
    Refmac fails if _entry.id is longer than 80 chars including quotations
    """
    st_new = st.clone()
    logger.writeln("Writing mmCIF file: {}".format(cif_out))
    if cif_ref:
        logger.writeln("  using mmCIF metadata from: {}".format(cif_ref))
        groups = gemmi.MmcifOutputGroups(False)
        groups.group_pdb = True
        groups.ncs = True
        groups.atoms = True
        groups.cell = True
        groups.scale = True
        groups.assembly = True
        groups.entity = True
        groups.entity_poly_seq = True
        groups.cis = True
        groups.conn = True
        # FIXME is this all? 
        try:
            doc = read_cif_safe(cif_ref)
        except Exception as e:
            # Sometimes refmac writes a broken mmcif file..
            logger.error("Error in mmCIF reading: {}".format(e))
            logger.error("  Give up using cif reference.")
            return write_mmcif(st, cif_out)
            
        blocks = list(filter(lambda b: b.find_loop("_atom_site.id"), doc))
        if len(blocks) == 0:
            logger.writeln("No _atom_site found in {}".format(cif_ref))
            logger.writeln("  Give up using cif reference.")
            return write_mmcif(st, cif_out)
        block = blocks[0]
        # to remove fract_transf_matrix. maybe we should keep some (like _atom_sites.solution_hydrogens)?
        # we do not want this because cell may be updated
        block.find_mmcif_category("_atom_sites.").erase()
        st_new.update_mmcif_block(block, groups)
        if "_entry.id" in st_new.info: st_new.info["_entry.id"] = st_new.info["_entry.id"][:78]
        doc.write_file(cif_out, style=gemmi.cif.Style.Aligned)
    else:
        st_new.name = st_new.name[:78] # this will become _entry.id
        if "_entry.id" in st_new.info: st_new.info["_entry.id"] = st_new.info["_entry.id"][:78]
        groups = gemmi.MmcifOutputGroups(True)
        doc = gemmi.cif.Document()
        block = doc.add_new_block("new")
        st_new.update_mmcif_block(block, groups)
        doc.write_file(cif_out, style=gemmi.cif.Style.Aligned)
# write_mmcif()

def write_pdb(st, pdb_out):
    logger.writeln("Writing PDB file: {}".format(pdb_out))
    chain_id_lens = [len(x) for x in model.all_chain_ids(st)]
    if chain_id_lens and max(chain_id_lens) > 2:
        st = st.clone()
        st.shorten_chain_names()
    st.write_pdb(pdb_out, use_linkr=True)
# write_pdb()

def write_model(st, prefix=None, file_name=None, pdb=False, cif=False, cif_ref=None, hout=True):
    if not hout and st[0].has_hydrogen():
        st = st.clone()
        st.remove_hydrogens()
    if file_name:
        if file_name.endswith("cif"):
            write_mmcif(st, file_name, cif_ref)
        elif file_name.endswith(".pdb"):
            write_pdb(st, file_name)
        else:
            raise Exception("Cannot determine file format from file name: {}".format(file_name))
    else:            
        if cif:
            write_mmcif(st, prefix+".mmcif", cif_ref)
        if pdb:
            write_pdb(st, prefix+".pdb")
# write_model()

def read_shifts_txt(shifts_txt):
    ret = {}
    s = open(shifts_txt).read()
    s = s.replace("\n"," ").split()
    for i in range(len(s)-3):
        if s[i] in ("pdbin", "pdbout") and s[i+1] in ("cell", "shifts"):
            n = 6 if s[i+1] == "cell" else 3
            ret["{} {}".format(s[i], s[i+1])] = [float(x) for x in s[i+2:i+2+n]]

    return ret
# read_shifts_txt()

def read_ccp4_map(filename, setup=True, default_value=0., pixel_size=None, ignore_origin=True):
    m = gemmi.read_ccp4_map(filename)
    g = m.grid
    grid_cell = [m.header_i32(x) for x in (8,9,10)]
    grid_start = [m.header_i32(x) for x in (5,6,7)]
    grid_shape = [m.header_i32(x) for x in (1,2,3)]
    axis_pos = m.axis_positions()
    axis_letters = ["","",""]
    for i, l in zip(axis_pos, "XYZ"): axis_letters[i] = l
    spacings = [1./g.unit_cell.reciprocal().parameters[i]/grid_cell[i] for i in (0,1,2)]
    voxel_size = [g.unit_cell.parameters[i]/grid_cell[i] for i in (0,1,2)]
    origin = [m.header_float(x) for x in (50,51,52)]
    label = m.header_str(57, 80)
    label = label[:label.find("\0")]
    logger.writeln("Reading CCP4/MRC map file {}".format(filename))
    logger.writeln("   Cell Grid: {:4d} {:4d} {:4d}".format(*grid_cell))
    logger.writeln("    Map mode: {}".format(m.header_i32(4)))
    logger.writeln("       Start: {:4d} {:4d} {:4d}".format(*grid_start))
    logger.writeln("       Shape: {:4d} {:4d} {:4d}".format(*grid_shape))
    logger.writeln("        Cell: {} {} {} {} {} {}".format(*g.unit_cell.parameters))
    logger.writeln("  Axis order: {}".format(" ".join(axis_letters)))
    logger.writeln(" Space group: {}".format(m.header_i32(23)))
    logger.writeln("     Spacing: {:.6f} {:.6f} {:.6f}".format(*spacings))
    logger.writeln("  Voxel size: {:.6f} {:.6f} {:.6f}".format(*voxel_size))
    logger.writeln("      Origin: {:.6e} {:.6e} {:.6e}".format(*origin))
    if not numpy.all(numpy.asarray(origin) == 0.):
        logger.writeln("             ! WARNING: ORIGIN header is not supported.")
        if ignore_origin:
            logger.writeln("             ! WARNING: removing ORIGIN values. This might cause a misalignment between map and model.")
            for i in (50,51,52): m.set_header_float(i, 0.)
    logger.writeln("       Label: {}".format(label))
    logger.writeln("")

    if setup:
        if default_value is None: default_value = float("nan")
        m.setup(default_value)
        grid_start = [grid_start[i] for i in axis_pos]
        
    if pixel_size is not None:
        try:
            len(pixel_size)
        except TypeError:
            pixel_size = [pixel_size, pixel_size, pixel_size]
            
        logger.writeln("Overriding pixel size with {:.6f} {:.6f} {:.6f}".format(*pixel_size))
        orgc = m.grid.unit_cell.parameters
        new_abc = [orgc[i]*pixel_size[i]/voxel_size[i] for i in (0,1,2)]
        m.grid.unit_cell = gemmi.UnitCell(new_abc[0], new_abc[1], new_abc[2],
                                          orgc[3], orgc[4], orgc[5])
        logger.writeln(" New cell= {:.1f} {:.1f} {:.1f} {:.1f} {:.1f} {:.1f}".format(*m.grid.unit_cell.parameters))

    return [m.grid, grid_start, grid_shape]
# read_ccp4_map()

def read_halfmaps(files, pixel_size=None, fail=True):
    if fail and len(files) != 2:
        raise SystemExit("Error: Give exactly two files for half maps")
    maps = [read_ccp4_map(f, pixel_size=pixel_size) for f in files]
    if numpy.array_equal(maps[0][0].array, maps[1][0].array):
        raise SystemExit("Error: Half maps have exactly the same values. Check your input.")
    
    assert maps[0][0].shape == maps[1][0].shape
    assert maps[0][0].unit_cell == maps[1][0].unit_cell
    assert maps[0][1] == maps[1][1]
    
    return maps
# read_halfmaps()

def read_mmhkl(hklin, cif_index=0): # mtz or mmcif
    spext = splitext(hklin)
    if spext[1].lower() == ".mtz":
        logger.writeln("Reading MTZ file: {}".format(hklin))
        mtz = gemmi.read_mtz_file(hklin)
    elif spext[1].lower() in (".cif", ".ent"):
        logger.writeln("Reading mmCIF file (hkl data): {} at index {}".format(hklin, cif_index+1))
        doc = gemmi.cif.read(hklin)
        blocks = gemmi.as_refln_blocks(doc)
        cif2mtz = gemmi.CifToMtz()
        mtz = cif2mtz.convert_block_to_mtz(blocks[cif_index])
    else:
        raise RuntimeError("Unsupported file type: {}".format(spext[1]))
    if mtz.spacegroup is None:
        raise RuntimeError("Missing space group information")
    logger.writeln("    Unit cell: {:.4f} {:.4f} {:.4f} {:.3f} {:.3f} {:.3f}".format(*mtz.cell.parameters))
    logger.writeln("  Space group: {}".format(mtz.spacegroup.xhm()))
    logger.writeln("      Columns: {}".format(" ".join(mtz.column_labels())))
    logger.writeln("")
    return mtz
# read_mmhkl()

def is_mmhkl_file(hklin):
    spext = splitext(hklin)
    if spext[1].lower() == ".mtz":
        return True
    if spext[1].lower() == ".hkl": # macromolecule files should not have .hkl extension
        return False
    if spext[1].lower() in (".cif", ".ent"):
        for b in gemmi.cif.read(hklin):
            if b.find_values("_refln.index_h"):
                return True
            if b.find_values("_refln_index_h"):
                return False
    # otherwise cannot decide
# is_smhkl()

def read_map_from_mtz(mtz_in, cols, grid_size=None, sample_rate=3):
    mtz = read_mmhkl(mtz_in)
    d_min = mtz.resolution_high() # TODO get resolution for column?
    if grid_size is None:
        grid_size = mtz.get_size_for_hkl(sample_rate=sample_rate)
    F = mtz.get_f_phi_on_grid(cols[0], cols[1], grid_size)
    m = gemmi.transform_f_phi_grid_to_map(F)
    return d_min, m
# read_map_from_mtz()

def read_asu_data_from_mtz(mtz_in, cols):
    assert 0 < len(cols) < 3
    mtz = read_mmhkl(mtz_in)
    sg = mtz.spacegroup
    miller = mtz.make_miller_array()
    f = mtz.column_with_label(cols[0])
    cell = mtz.get_cell(f.dataset_id)
    if len(cols) == 2:
        phi = mtz.column_with_label(cols[1])
        assert f.type == "F"
        assert phi.type == "P"
        phi = numpy.deg2rad(phi)
        f_comp = f * (numpy.cos(phi) + 1j * numpy.sin(phi))
        asu = gemmi.ComplexAsuData(cell, sg, miller, f_comp) # ensure asu?
        return asu
    else:
        if f.is_integer():
            gr_t = gemmi.IntAsuData
        else:
            gr_t = gemmi.FloatAsuData
        
        asu = gr_t(cell, sg, miller, f)
        return asu
# read_asu_data_from_mtz()

def read_cif_safe(cif_in):
    ifs = gzip.open(cif_in, "rt") if cif_in.endswith(".gz") else open(cif_in)
    s = ifs.read()
    if "\0" in s: # Refmac occasionally writes \0 in some fields..
        logger.writeln(" WARNING: null character detected. Replacing with '.'")
        s = s.replace("\0", ".")
    doc = gemmi.cif.read_string(s)
    return doc
# read_cif_safe()

def read_structure(xyz_in, assign_het_flags=True):
    spext = splitext(xyz_in)
    st = None
    if spext[1].lower() in (".pdb", ".ent"):
        logger.writeln("Reading PDB file: {}".format(xyz_in))
        st = gemmi.read_pdb(xyz_in)
    elif spext[1].lower() in (".cif", ".mmcif"):
        doc = read_cif_safe(xyz_in)
        for block in doc:
            if block.find_loop("_atom_site.id"):
                if st is None:
                    logger.writeln("Reading mmCIF file: {}".format(xyz_in))
                    st =  gemmi.make_structure_from_block(block)
                else:
                    logger.writeln(" WARNING: more than one block having structure found. Will use first one.")
                    break
            elif block.find_loop("_atom_site_label"):
                if st is None:
                    logger.writeln("Reading smCIF file: {}".format(xyz_in))
                    ss = gemmi.read_small_structure(xyz_in)
                    if not ss.sites:
                        raise RuntimeError("No atoms found in cif file.")
                    st = model.cx_to_mx(ss)
                else:
                    logger.writeln(" WARNING: more than one block having structure found. Will use first one.")
                    break
            elif (block.find_loop("_chem_comp_atom.x") or
                  block.find_loop("_chem_comp_atom.model_Cartn_x") or
                  block.find_loop("_chem_comp_atom.pdbx_model_Cartn_x_ideal")):
                if st is None:
                    logger.writeln("Reading chemical component file: {}".format(xyz_in))
                    st = gemmi.make_structure_from_chemcomp_block(block)
    elif spext[1].lower() in (".ins", ".res"):
        logger.writeln("Reading SHELX ins/res file: {}".format(xyz_in))
        st = model.cx_to_mx(read_shelx_ins(ins_in=xyz_in)[0])
        st.setup_cell_images()
    else:
        raise RuntimeError("Unsupported file type: {}".format(spext[1]))
    if st is not None:
        if st.cell.is_crystal():
            logger.writeln("    Unit cell: {:.4f} {:.4f} {:.4f} {:.3f} {:.3f} {:.3f}".format(*st.cell.parameters))
            logger.writeln("  Space group: {}".format(st.spacegroup_hm))
        if st.ncs:
            n_given = sum(1 for x in st.ncs if x.given)
            logger.writeln(" No. strict NCS: {} ({} already applied)".format(len(st.ncs), n_given))
        logger.writeln("")
    if assign_het_flags:
        st.assign_het_flags()
    return st
# read_structure()

def read_structure_from_pdb_and_mmcif(xyz_in):
    st = read_structure(xyz_in)
    cif_ref = None
    spext = splitext(xyz_in)
    if spext[1] in (".pdb", ".ent"):
        cif_in = spext[0] + ".mmcif"
        if os.path.isfile(cif_in):
            print(" Will use mmcif metadata from {}".format(cif_in))
            cif_ref = cif_in
    elif spext[1] in (".cif", ".mmcif"):
        cif_ref = xyz_in
        pdb_in = spext[0] + ".pdb"
        if os.path.isfile(pdb_in):
            print(" Reading PDB REMARKS from {}".format(pdb_in))
            tmp = gemmi.read_structure(pdb_in)
            st.raw_remarks = tmp.raw_remarks

    if cif_ref is None and xyz_in.endswith("cif"):
        cif_ref = xyz_in
            
    return st, cif_ref
# read_structure_from_pdb_and_mmcif()

def merge_ligand_cif(cifs_in, cif_out):
    docs = [gemmi.cif.read(x) for x in cifs_in]
    tags = dict(comp=["_chem_comp.id"],
                link=["_chem_link.id"],
                mod=["_chem_mod.id"])
    list_names = [k+"_list" for k in tags]

    # Check duplicated block names
    names = {}
    for i, doc in enumerate(docs):
        for j, b in enumerate(doc):
            if b.name not in list_names and not b.name.startswith("mod_"):
                names.setdefault(b.name, []).append((i,j))

    # Keep only last one if duplicated
    todel = []
    for k in names:
        if len(names[k]) > 1:
            for i,j in reversed(names[k][:-1]):
                logger.writeln("WARNING: removing duplicated {} from {}".format(k, cifs_in[i]))
                todel.append((i,j))
                for t in "comp", "link":
                    if k.startswith("{}_".format(t)):
                        comp_list = docs[i].find_block("{}_list".format(t))
                        table = comp_list.find("_chem_{}.".format(t), ["id"])
                        for l in reversed([l for l, row in enumerate(table) if row.str(0) == k[5:]]):
                            table.remove_row(l)

    for i,j in sorted(todel, reverse=True):
        del docs[i][j]

    # Accumulate list
    found = dict(comp=0, link=0, mod=0)
    for d in docs:
        for k in tags:
            b = d.find_block("{}_list".format(k))
            if not b: continue
            found[k] += 1
            l = b.find_loop(tags[k][0]).get_loop()
            for t in l.tags:
                if t not in tags[k]: tags[k].append(t)

    # Check duplicated modifications
    known_mods = [] # need to check monomer library?
    for d in docs:
        restraints.rename_cif_modification_if_necessary(d, known_mods)
        mod_list = d.find_block("mod_list")
        if mod_list:
            for row in mod_list.find("_chem_mod.", ["id"]):
                known_mods.append(row.str(0))
        
    doc = gemmi.cif.Document()
    # Add lists
    for k in tags:
        if not found[k]: continue
        lst = doc.add_new_block("{}_list".format(k))
        loop = lst.init_loop("", tags[k])
        tags_for_find = [tags[k][0]] + ["?"+x for x in tags[k][1:]]
        
        for d in docs:
            b = d.find_block("{}_list".format(k))
            if not b: continue
            vals = b.find(tags_for_find)
            for v in vals:
                rl = [v.get(x) if v.has(x) else "." for x in range(len(tags[k]))]
                loop.add_row(rl)

    # Add other items
    for d in docs:
        for b in d:
            if b.name not in list_names:
                doc.add_copied_block(b)

    doc.write_file(cif_out, style=gemmi.cif.Style.Aligned)
# merge_ligand_cif()

def read_shelx_ins(ins_in=None, lines_in=None, ignore_q_peaks=True): # TODO support gz?
    assert (ins_in, lines_in).count(None) == 1
    ss = gemmi.SmallStructure()

    keywords = """
    TITL CELL ZERR LATT SYMM SFAC NEUT DISP UNIT LAUE REM  MORE END  HKLF OMIT SHEL BASF TWIN TWST EXTI SWAT
    ABIN ANSC ANSR MERG LIST SPEC RESI MOVE ANIS AFIX HFIX FRAG FEND EXYZ EADP EQIV CONN PART BIND FREE DFIX
    DANG BUMP SAME SADI CHIV FLAT DELU SIMU RIGU PRIG DEFS ISOR XNPD NCSY SUMP L.S. CGLS BLOC DAMP STIR WGHT
    FVAR WIGL BOND CONF MPLA RTAB HTAB ACTA SIZE TEMP WPDB FMAP GRID PLAN TIME HOPE MOLE
    """.split()
    re_kwd = re.compile("^({})_?".format("|".join(keywords)))
    
    # remove comments/blanks and concatenate lines
    lines = []
    concat_flag = False
    for l in open(ins_in) if ins_in else lines_in:
        l = l.rstrip()
        if l.startswith("REM"): continue
        if l.startswith(";"): continue
        if not l.strip(): continue

        if l.endswith("="):
            l = l[:l.rindex("=")]
            if concat_flag:
                lines[-1] += l
            else:
                lines.append(l)
            concat_flag = True
        elif concat_flag:
            lines[-1] += l
            concat_flag = False
        else:
            lines.append(l)

    # parse lines
    sfacs = []
    latt, symms = 1, []
    info = dict(hklf=0)
    for l in lines:
        sp = l.split()
        ins = sp[0].upper()
        if ins == "TITL":
            pass
        elif l.startswith(" "): # title continued? instructions after space is allowed??
            pass
        elif ins == "CELL":
            #ss.wavelength = float(sp[1]) # next gemmi ver.
            ss.cell.set(*map(float, sp[2:]))
        elif ins == "LATT":
            latt = int(sp[1])
        elif ins == "SYMM":
            symms.append(gemmi.Op("".join(sp[1:])).wrap())
        elif ins == "SFAC": # TODO check numbers?
            if len(sp) < 2: continue
            sfacs.append(gemmi.Element(sp[1]))
            if len(sp) > 2:
                try: float(sp[2])
                except ValueError:
                    sfacs.extend([gemmi.Element(x) for x in sp[2:]])
        elif ins == "HKLF":
            info["hklf"] = int(sp[1])
        elif not re_kwd.search(ins):
            if not 4 < len(sp) < 13:
                logger.writeln("cannot parse this line: {}".format(l))
                continue
            site = gemmi.SmallStructure.Site()
            site.label = sp[0]
            try:
                site.element = sfacs[int(sp[1])-1]
            except:
                logger.error("failed to parse: {}".format(l))
                continue

            if site.label.startswith("Q") and ignore_q_peaks:
                logger.writeln("skip Q peak: {}".format(l))
                continue
            
            site.fract.fromlist(list(map(float, sp[2:5])))
            if len(sp) > 5:
                q = abs(float(sp[5]))
                if q > 10: q = q % 10 # FIXME proper handling
                site.occ = q
            if len(sp) > 11:
                u = list(map(float, sp[6:12]))
                site.aniso = gemmi.SMat33d(u[0], u[1], u[2], u[5], u[4], u[3])
                #TODO site.u_iso needs to be set?
            else:
                site.u_iso = float(sp[6])

            ss.add_site(site)

    # Determine space group
    if gemmi.Op() not in symms: # identity operator may not be present in ins file
        symms.append(gemmi.Op())

    lops = {1: [], # P
            2: [gemmi.Op("x+1/2,y+1/2,z+1/2")], # I
            3: [gemmi.Op("x+2/3,y+1/3,z+1/3"),  # R
                gemmi.Op("x+1/3,y+2/3,z+2/3")],
            4: [gemmi.Op("x,y+1/2,z+1/2"),      # F
                gemmi.Op("x+1/2,y,z+1/2"),
                gemmi.Op("x+1/2,y+1/2,z")],
            5: [gemmi.Op("x,y+1/2,z+1/2")],     # A
            6: [gemmi.Op("x+1/2,y,z+1/2")],     # B
            7: [gemmi.Op("x+1/2,y+1/2,z")],     # C
            }
    for op in lops[abs(latt)]:
        symms.extend([x*op for x in symms])
    if latt > 0:
        symms.extend([x*gemmi.Op("-x,-y,-z") for x in symms])

    ss.symops = [op.triplet() for op in set(symms)]
    ss.set_spacegroup("s")
    # in case of non-regular setting, gemmi.SpaceGroup cannot be constructed anyway.
    if ss.spacegroup is None:
        raise RuntimeError("Cannot construct space group from symbols: {}".format(ss.symops))
    return ss, info
# read_shelx_ins()

def read_shelx_hkl(cell, sg, hklf, file_in=None, lines_in=None):
    assert (file_in, lines_in).count(None) == 1
    hkls, vals, sigs = [], [], []
    for l in open(file_in) if file_in else lines_in:
        if l.startswith(";"): continue
        if not l.strip() or len(l) < 25: continue
        try:
            hkl = int(l[:4]), int(l[4:8]), int(l[8:12])
        except ValueError:
            logger.writeln("Error while parsing HKL part: {}".format(l))
            break
            
        if hkl == (0,0,0): break
        hkls.append(hkl)
        vals.append(float(l[12:20]))
        sigs.append(float(l[20:28]))
        # batch = l[28:32]
        # wavelength = l[32:40]

    ints = gemmi.Intensities()
    ints.set_data(cell, sg, hkls, vals, sigs)
    ints.merge_in_place(gemmi.DataType.Anomalous)
    if not (ints.isign_array < 0).any(): ints.type = gemmi.DataType.Mean
    logger.writeln(" Multiplicity: max= {} mean= {:.1f} min= {}".format(numpy.max(ints.nobs_array),
                                                                     numpy.mean(ints.nobs_array),
                                                                     numpy.min(ints.nobs_array)))
    mtz = ints.prepare_merged_mtz(with_nobs=False)
    if hklf == 3:
        conv = {"IMEAN": ("FP", "F"),
                "SIGIMEAN": ("SIGFP", "Q"),
                "I(+)": ("F(+)", "G"),
                "SIGI(+)": ("SIGF(+)", "L"),
                "I(-)": ("F(-)", "G"),
                "SIGI(-)": ("SIGF(-)", "L"),
                }
        for col in mtz.columns:
            if col.label in conv:
                col.label, col.type = conv[col.label]
    return mtz
# read_shelx_hkl()

def read_smcif_hkl(cif_in, cell_if_absent=None, sg_if_absent=None):
    # Very crude support for smcif - just because I do not know other varieties.
    # TODO other possible data types? (amplitudes?)
    # TODO check _refln_observed_status?
    logger.writeln("Reading hkl data from smcif: {}".format(cif_in))
    b = gemmi.cif.read(cif_in).sole_block()
    try:
        cell_par = [float(b.find_value("_cell_length_{}".format(x))) for x in ("a", "b", "c")]
        cell_par += [float(b.find_value("_cell_angle_{}".format(x))) for x in ("alpha", "beta", "gamma")]
        cell = gemmi.UnitCell(*cell_par)
        logger.writeln("    Unit cell: {:.4f} {:.4f} {:.4f} {:.3f} {:.3f} {:.3f}".format(*cell.parameters))
    except:
        logger.writeln(" WARNING: no unit cell in this file")
        cell = cell_if_absent

    for optag in ("_space_group_symop_operation_xyz", "_symmetry_equiv_pos_as_xyz"):
        ops = [gemmi.Op(gemmi.cif.as_string(x)) for x in b.find_loop(optag)]
        sg = gemmi.find_spacegroup_by_ops(gemmi.GroupOps(ops))
        if sg:
            logger.writeln("  Space group: {}".format(sg.xhm()))
            break
    else:
        sg = sg_if_absent

    if cell is None or sg is None:
        raise RuntimeError("Cell and/or symmetry operations not found in {}".format(cif_in))
        
    l = b.find_values("_refln_index_h").get_loop()
    i_hkl = [l.tags.index("_refln_index_{}".format(h)) for h in "hkl"]
    i_int = l.tags.index("_refln_F_squared_meas")
    i_sig = l.tags.index("_refln_F_squared_sigma")
    hkls, vals, sigs = [], [], []
    for i in range(l.length()):
        hkl = [gemmi.cif.as_int(l[i, j]) for j in i_hkl]
        hkls.append(hkl)
        vals.append(gemmi.cif.as_number(l[i, i_int]))
        sigs.append(gemmi.cif.as_number(l[i, i_sig]))
    
    ints = gemmi.Intensities()
    ints.set_data(cell, sg, hkls, vals, sigs)
    ints.merge_in_place(gemmi.DataType.Anomalous)
    if not (ints.isign_array < 0).any(): ints.type = gemmi.DataType.Mean
    logger.writeln(" Multiplicity: max= {} mean= {:.1f} min= {}".format(numpy.max(ints.nobs_array),
                                                                     numpy.mean(ints.nobs_array),
                                                                     numpy.min(ints.nobs_array)))
    logger.writeln("")
    return ints.prepare_merged_mtz(with_nobs=False)
# read_smcif_hkl()
    
def read_smcif_shelx(cif_in):
    logger.writeln("Reading small molecule cif: {}".format(cif_in))
    b = gemmi.cif.read(cif_in).sole_block()
    res_str = b.find_value("_shelx_res_file")
    hkl_str = b.find_value("_shelx_hkl_file")
    if not res_str: raise RuntimeError("_shelx_res_file not found in {}".format(cif_in))
    if not hkl_str: raise RuntimeError("_shelx_hkl_file not found in {}".format(cif_in))
    
    ss, info = read_shelx_ins(lines_in=res_str.splitlines())
    mtz = read_shelx_hkl(ss.cell, ss.spacegroup, info.get("hklf"), lines_in=hkl_str.splitlines())
    return mtz, ss, info
# read_smcif_shelx()

def read_small_molecule_files(files):
    st, mtz, hklf = None, None, None
    # first pass - find structure
    for filename in files:
        ext = splitext(filename)[1]
        if ext in (".cif", ".res", ".ins"):
            try:
                st = read_structure(filename)
            except:
                continue
            logger.writeln("Coordinates read from: {}".format(filename))
            if ext == ".cif":
                b = gemmi.cif.read(filename).sole_block()
                res_str = b.find_value("_shelx_res_file")
            else:
                res_str = open(filename).read()
            if res_str:
                _, info = read_shelx_ins(lines_in=res_str.splitlines())
                hklf = info["hklf"]
    if st is None:
        logger.writeln("ERROR: coordinates not found.")
        return None, None
    
    # second pass - find hkl
    for filename in files:
        ext = splitext(filename)[1]
        try:
            b = gemmi.cif.read(filename).sole_block()
            hkl_str = b.find_value("_shelx_hkl_file")
            if hkl_str:
                mtz = read_shelx_hkl(st.cell, st.find_spacegroup(), hklf, lines_in=hkl_str.splitlines())
                logger.writeln("reflection data read from: {}".format(filename))
            elif b.find_loop("_refln_index_h"):
                mtz = read_smcif_hkl(filename, st.cell, st.find_spacegroup())
        except RuntimeError: # not a cif file
            if ext == ".hkl":
                mtz = read_shelx_hkl(st.cell, st.find_spacegroup(), hklf, file_in=filename)
                logger.writeln("reflection data read from: {}".format(filename))

    return st, mtz

def read_sequence_file(f):
    # TODO needs improvement
    # return a list of [name, sequence]
    ret = []
    for l in open(f):
        l = l.strip()
        if l.startswith(">"):
            name = l[1:].strip()
            ret.append([name, ""])
        elif l:
            if not ret: ret.append(["", ""])
            ret[-1][1] += l.replace("*", "").replace("-", "").upper()
    return ret



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import argparse
import os
import gemmi
import numpy
import json
import servalcat # for version
from servalcat.utils import logger
from servalcat import utils
from servalcat.refine.refine import Geom, Refine, convert_stats_to_dicts
from servalcat.refmac import refmac_keywords

def add_arguments(parser):
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--model',
                        help='Input atomic model file')
    group.add_argument('--update_dictionary', 
                       help="Dictionary file to be updated")
    parser.add_argument("--monlib",
                        help="Monomer library path. Default: $CLIBD_MON")
    parser.add_argument('--ligand', nargs="*", action="append",
                        help="restraint dictionary cif file(s)")
    parser.add_argument('--ncycle', type=int, default=10,
                        help="number of CG cycles (default: %(default)d)")
    parser.add_argument('--hydrogen', default="all", choices=["all", "yes", "no"],
                        help="all: add riding hydrogen atoms, yes: use hydrogen atoms if present, no: remove hydrogen atoms in input. "
                        "Default: %(default)s")
    parser.add_argument('--find_links', action='store_true', 
                        help='Automatically add links')
    parser.add_argument('--randomize', type=float, default=0,
                        help='Shake coordinates with specified rmsd')
    parser.add_argument('--ncsr', action='store_true', 
                        help='Use local NCS restraints')
    parser.add_argument('--keywords', nargs='+', action="append",
                        help="refmac keyword(s)")
    parser.add_argument('--keyword_file', nargs='+', action="append",
                        help="refmac keyword file(s)")
    parser.add_argument('-o','--output_prefix',
                        help="Output prefix")

# add_arguments()

def parse_args(arg_list):
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    return parser.parse_args(arg_list)
# parse_args()

def add_program_info_to_dictionary(block, comp_id, program_name="servalcat", descriptor="optimization tool"):
    # old acedrg used _pdbx_chem_comp_description_generator. and descriptor
    # new acedrg (>280?) uses _acedrg_chem_comp_descriptor. and type
    for tag, name in (("_acedrg_chem_comp_descriptor.", "type"),
                      ("_pdbx_chem_comp_description_generator.", "descriptor")):
        tab = block.find(tag, ["program_name", "program_version", name])
        if tab:
            loop = tab.loop
            # just overwrite version if it's there
            for row in tab:
                if row.str(0) == program_name and row.str(2) == descriptor:
                    row[1] = gemmi.cif.quote(servalcat.__version__)
                    return
            break
    else:
        # it may be strange to say _acedrg in this case..
        name = "type"
        loop = block.init_loop("_acedrg_chem_comp_descriptor.", ["comp_id",
                                                                 "program_name",
                                                                 "program_version",
                                                                 name])
    tags = [x[x.index(".")+1:] for x in loop.tags]
    row = ["" for _ in range(len(tags))]
    for tag, val in (("comp_id", comp_id),
                     ("program_name", program_name),
                     ("program_version", servalcat.__version__),
                     (name, descriptor)):
        if tag in tags: row[tags.index(tag)] = val
    loop.add_row(gemmi.cif.quote_list(row))
# add_program_info_to_dictionary()

def refine_and_update_dictionary(cif_in, monomer_dir, output_prefix, randomize=0, ncycle1=10, ncycle2=30):
    doc = gemmi.cif.read(cif_in)
    for block in doc: # this block will be reused below
        st = gemmi.make_structure_from_chemcomp_block(block)
        if len(st) > 0: break
    else:
        raise SystemExit("No model in the cif file")
    monlib = utils.restraints.load_monomer_library(st, monomer_dir=monomer_dir, # monlib is needed for ener_lib
                                                   cif_files=[cif_in],
                                                   stop_for_unknowns=True)
    all_stats = []
    for i_macro in 0, 1:
        try:
            topo, _ = utils.restraints.prepare_topology(st, monlib, h_change=[gemmi.HydrogenChange.Remove, gemmi.HydrogenChange.ReAdd][i_macro],
                                                        check_hydrogen=(i_macro == 1))
        except RuntimeError as e:
            raise SystemExit("Error: {}".format(e))

        geom = Geom(st, topo, monlib, shake_rms=randomize)
        refiner = Refine(st, geom)
        logger.writeln("Running {} cycles with wchir=4 wvdw=2 {} hydrogen".format(ncycle1, ["without","with"][i_macro]))
        geom.calc_kwds["wchir"] = 4
        geom.calc_kwds["wvdw"] = 2
        all_stats.append(refiner.run_cycles(ncycle1))

        logger.writeln("Running {} cycles with wchir=1 wvdw=2 {} hydrogen".format(ncycle2, ["without","with"][i_macro]))
        geom.calc_kwds["wchir"] = 1
        geom.calc_kwds["wvdw"] = 2
        all_stats.append(refiner.run_cycles(ncycle2))

    # replace xyz
    pos = {cra.atom.name: cra.atom.pos.tolist() for cra in refiner.st[0].all()}
    for row in block.find("_chem_comp_atom.", ["atom_id", "?x", "?y", "?z",
                                               "?pdbx_model_Cartn_x_ideal",
                                               "?pdbx_model_Cartn_y_ideal",
                                               "?pdbx_model_Cartn_z_ideal"]):
        p = pos[row.str(0)]
        for i in range(3):
            if row.has(i+1):
                row[i+1] = "{:.3f}".format(p[i])
            if row.has(i+4):
                row[i+4] = "{:.3f}".format(p[i])
    # add description
    add_program_info_to_dictionary(block, st[0][0][0].name)
    doc.write_file(output_prefix + "_updated.cif", style=gemmi.cif.Style.Aligned)
    logger.writeln("Updated dictionary saved: {}".format(output_prefix + "_updated.cif"))
    with open(output_prefix + "_stats.json", "w") as ofs:
        json.dump([convert_stats_to_dicts(x) for x in all_stats],
                  ofs, indent=2)
        logger.writeln("Refinement statistics saved: {}".format(ofs.name))
# refine_and_update_dictionary()

def refine_geom(model_in, monomer_dir, cif_files, h_change, ncycle, output_prefix, randomize, params,
                find_links=False, use_ncsr=False):
    st = utils.fileio.read_structure(model_in)
    utils.model.setup_entities(st, clear=True, force_subchain_names=True, overwrite_entity_type=True)
    if not all(op.given for op in st.ncs):
        st2 = st.clone()
        logger.writeln("Take NCS constraints into account.")
        st2.expand_ncs(gemmi.HowToNameCopiedChain.Dup, merge_dist=0)
        utils.fileio.write_model(st2, file_name="input_expanded.pdb")

    monlib = utils.restraints.load_monomer_library(st, monomer_dir=monomer_dir,
                                                   cif_files=cif_files,
                                                   stop_for_unknowns=True,
                                                   params=params)
    utils.restraints.find_and_fix_links(st, monlib, add_found=find_links) # should remove unknown id here?
    try:
        topo, _ = utils.restraints.prepare_topology(st, monlib, h_change=h_change,
                                                    check_hydrogen=(h_change==gemmi.HydrogenChange.NoChange),
                                                    params=params)
    except RuntimeError as e:
        raise SystemExit("Error: {}".format(e))

    if use_ncsr:
        ncslist = utils.restraints.prepare_ncs_restraints(st)
    else:
        ncslist = False
    geom = Geom(st, topo, monlib, shake_rms=randomize, params=params, ncslist=ncslist)
    refiner = Refine(st, geom)
    stats = refiner.run_cycles(ncycle,
                               stats_json_out=output_prefix + "_stats.json")
    refiner.st.name = output_prefix
    utils.fileio.write_model(refiner.st, output_prefix, pdb=True, cif=True)
# refine_geom()

def main(args):
    keywords = []
    if args.keywords: keywords = sum(args.keywords, [])
    if args.keyword_file: keywords.extend(l for f in sum(args.keyword_file, []) for l in open(f))
    decide_prefix = lambda f: utils.fileio.splitext(os.path.basename(f))[0] + "_refined"
    if args.model:
        params = refmac_keywords.parse_keywords(keywords)
        if not args.output_prefix:
            args.output_prefix = decide_prefix(args.model)
        if args.ligand:
            args.ligand = sum(args.ligand, [])
        h_change = {"all":gemmi.HydrogenChange.ReAddButWater,
                    "full":gemmi.HydrogenChange.ReAdd,
                    "yes":gemmi.HydrogenChange.NoChange,
                    "no":gemmi.HydrogenChange.Remove}[args.hydrogen]
        refine_geom(model_in=args.model,
                    monomer_dir=args.monlib,
                    cif_files=args.ligand,
                    h_change=h_change,
                    ncycle=args.ncycle,
                    output_prefix=args.output_prefix,
                    randomize=args.randomize,
                    params=params,
                    find_links=args.find_links,
                    use_ncsr=args.ncsr)
    else:
        if not args.output_prefix:
            args.output_prefix = decide_prefix(args.update_dictionary)
        if args.ligand:
            logger.writeln("WARNING: monlib and ligand are ignored in the dictionary updating mode")
        if keywords:
            logger.writeln("WARNING: refmac keywords are ignored in the dictionary updating mode")
        refine_and_update_dictionary(cif_in=args.update_dictionary,
                                     monomer_dir=args.monlib,
                                     output_prefix=args.output_prefix,
                                     randomize=args.randomize)
# main()

if __name__ == "__main__":
    import sys
    args = parse_args(sys.argv[1:])
    main(args)



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
from servalcat.utils import logger
from servalcat.refmac import refmac_keywords
from servalcat import ext
import os
import io
import gemmi
import string
import random
import numpy
import pandas
import json
import fnmatch

default_proton_scale = 1.13 # scale of X-proton distance to X-H(e) distance

def decide_new_mod_id(mod_id, mods):
    # Refmac only allows up to 8 characters
    letters = string.digits + string.ascii_lowercase
    if len(mod_id) < 8:
        for l in letters:
            new_id = "{}{}{}".format(mod_id, "" if len(mod_id)==7 else "-", l)
            if new_id not in mods:
                return new_id

    # give up keeping original name
    while True: # XXX risk of infinite loop.. less likely though
        new_id = "mod" + "".join([random.choice(letters) for _ in range(4)])
        if new_id not in mods:
            return new_id
# decide_new_mod_id()

def rename_cif_modification_if_necessary(doc, known_ids):
    # FIXME Problematic if other file refers to modification that is renamed in this function - but how can we know?
    trans = {}
    for b in doc:
        for row in b.find("_chem_mod.", ["id"]):
            mod_id = row.str(0)
            if mod_id in known_ids:
                new_id = decide_new_mod_id(mod_id, known_ids)
                trans[mod_id] = new_id
                row[0] = new_id # modify id
                logger.writeln("INFO:: renaming modification id {} to {}".format(mod_id, new_id))

    # modify ids in mod_* blocks
    for mod_id in trans:
        b = doc.find_block("mod_{}".format(mod_id))
        if not b: # should raise error?
            logger.writeln("WARNING:: inconsistent mod description for {}".format(mod_id))
            continue
        b.name = "mod_{}".format(trans[mod_id]) # modify name
        for item in b:
            for tag in item.loop.tags:
                if tag.endswith(".mod_id"):
                    for row in b.find(tag[:tag.rindex(".")+1], ["mod_id"]):
                        row[0] = trans[mod_id]

    # Update mod id in links
    if trans:
        for b in doc:
            for row in b.find("_chem_link.", ["mod_id_1", "mod_id_2"]):
                for i in range(2):
                    if row.str(i) in trans:
                        row[i] = trans[row.str(i)]

    return trans
# rename_cif_modification_if_necessary()

def load_monomer_library(st, monomer_dir=None, cif_files=None, stop_for_unknowns=False,
                         ignore_monomer_dir=False, update_old_atom_names=True,
                         params=None):
    resnames = st[0].get_all_residue_names()

    if monomer_dir is None and not ignore_monomer_dir:
        if "CLIBD_MON" not in os.environ:
            logger.error("WARNING: CLIBD_MON is not set")
        else:
            monomer_dir = os.environ["CLIBD_MON"]

    if cif_files is None:
        cif_files = []
        
    if monomer_dir and not ignore_monomer_dir:
        if not os.path.isdir(monomer_dir):
            raise RuntimeError("not a directory: {}".format(monomer_dir))

        logger.writeln("Reading monomers from {}".format(monomer_dir))
        monlib = gemmi.read_monomer_lib(monomer_dir, resnames, ignore_missing=True)
    else:
        monlib = gemmi.MonLib()

    for f in cif_files:
        logger.writeln("Reading monomer: {}".format(f))
        doc = gemmi.cif.read(f)
        for b in doc:
            atom_id_list = b.find_values("_chem_comp_atom.atom_id")
            if atom_id_list:
                name = b.name.replace("comp_", "")
                if name in monlib.monomers:
                    logger.writeln("WARNING:: updating monomer {} using {}".format(name, f))
                    del monlib.monomers[name]

                # Check if bond length values are included
                # This is to fail if cif file is e.g. from PDB website
                if len(atom_id_list) > 1 and not b.find_values("_chem_comp_bond.value_dist"):
                    raise RuntimeError("{} does not contain bond length value for {}. You need to generate restraints (e.g. using acedrg).".format(f, name))
                    
            for row in b.find("_chem_link.", ["id"]):
                link_id = row.str(0)
                if link_id in monlib.links:
                    logger.writeln("WARNING:: updating link {} using {}".format(link_id, f))
                    del monlib.links[link_id]

        # If modification id is duplicated, need to rename
        rename_cif_modification_if_necessary(doc, monlib.modifications)
        monlib.read_monomer_doc(doc)
        for b in doc:
            for row in b.find("_chem_comp.", ["id", "group"]):
                if row.str(0) in monlib.monomers:
                    monlib.monomers[row.str(0)].set_group(row.str(1))

    not_loaded = set(resnames).difference(monlib.monomers)
    if not_loaded:
        logger.writeln("WARNING: monomers not loaded: {}".format(" ".join(not_loaded)))
        
    logger.writeln("Monomer library loaded: {} monomers, {} links, {} modifications".format(len(monlib.monomers),
                                                                                          len(monlib.links),
                                                                                          len(monlib.modifications)))
    logger.writeln("       loaded monomers: {}".format(" ".join([x for x in monlib.monomers])))
    logger.writeln("")

    logger.writeln("Checking if unknown atoms exist..")

    unknown_cc = set()
    for chain in st[0]: unknown_cc.update(res.name for res in chain if res.name not in monlib.monomers)
    if unknown_cc:
        if stop_for_unknowns:
            raise RuntimeError("Provide restraint cif file(s) for {}".format(",".join(unknown_cc)))
        else:
            logger.writeln("WARNING: ad-hoc restraints will be generated for {}".format(",".join(unknown_cc)))
            logger.writeln("         it is strongly recommended to generate them using AceDRG.")

    if update_old_atom_names:
        logger.write(monlib.update_old_atom_names(st))

    if params:
        update_torsions(monlib, params.get("restr", {}).get("torsion_include", {}))
    
    return monlib
# load_monomer_library()

def fix_elements_in_model(monlib, st):
    monlib_els = {m: {a.id: a.el for a in monlib.monomers[m].atoms} for m in monlib.monomers}
    for chain in st[0]:
        for res in chain:
            d = monlib_els.get(res.name)
            if not d: continue # should not happen
            for at in res:
                if at.name not in d: # for example atom names of element D may be different, which will be sorted later
                    continue
                el = d[at.name]
                if at.element != el:
                    logger.writeln(f"WARNING: correcting element of {st[0].get_cra(at)} to {el.name}")
                    at.element = el
# correct_elements_in_model()

def update_torsions(monlib, params):
    # take subset
    params = [p for p in params
              if any(x in p for x in ("tors_value", "tors_sigma", "tors_period"))]
    if not params:
        return
    logger.writeln("Updating torsion targets in dictionaries")
    for p in params:
        if "residue" in p:
            tors = [cc.rt.torsions for cc in monlib.monomers.values()
                    if fnmatch.fnmatch(cc.name, p["residue"])]
        elif "group" in p:
            g = gemmi.ChemComp.read_group(p["group"])
            # should warn if g is Null
            tors = [cc.rt.torsions for cc in monlib.monomers.values()
                    if cc.group == g]
        elif "link" in p:
            tors = [ln.rt.torsions for ln in monlib.links.values()
                    if fnmatch.fnmatch(ln.id, p["link"])]
        else:
            tors = []
        if not tors:
            continue
        logger.writeln(f" rule = {p}")
        for tt in tors:
            for t in tt:
                if fnmatch.fnmatch(t.label, p["tors_name"]):
                    if "tors_value" in p:
                        t.value = p["tors_value"]
                    if "tors_sigma" in p:
                        t.esd = p["tors_sigma"]
                    if "tors_period" in p:
                        t.period = p["tors_period"]
# update_torsions()

def make_torsion_rules(restr_params):
    # Defaults
    include_rules = [{"group": "peptide", "tors_name": "chi*"},
                     {"link": "*", "tors_name": "omega"},
                     {"residue": "*", "tors_name": "sp2_sp2*"},
                     {"link": "*", "tors_name": "sp2_sp2*"},
                     ]
    exclude_rules = []

    # Override include/exclude rules
    for i, name in enumerate(("torsion_include", "torsion_exclude")):
        rules = (include_rules, exclude_rules)[i]
        for p in restr_params.get(name, []):
            r = {}
            if p["flag"]:
                for k in "residue", "group", "link":
                    if k in p:
                        r[k] = p[k]
                if r and "tors_name" in p:
                    r["tors_name"] = p["tors_name"]
                    rules.append(r)
            else:
                rules.clear()

    # How to tell about hydrogen?
    logger.writeln("Torsion angle rules:")
    for l, rr in (("include", include_rules), ("exclude", exclude_rules)):
        logger.writeln(f" {l}:")
        if not rr:
            logger.writeln(f"  none")
        for r in rr:
            logger.writeln(f"  {r}")

    return include_rules, exclude_rules
# make_torsion_rules())

def select_restrained_torsions(monlib, include_rules, exclude_rules):
    ret = {"monomer": {}, "link": {}}
    
    # Collect monomer/link related torsions
    all_tors = {"mon": {}, "link": {}}
    groups = {}
    for mon_id in monlib.monomers:
        mon = monlib.monomers[mon_id]
        groups.setdefault(mon.group, []).append(mon_id)
        all_tors["mon"][mon_id] = [x.label for x in mon.rt.torsions]
    for mod_id in monlib.modifications:
        mod = monlib.modifications[mod_id]
        tors = [x.label for x in mod.rt.torsions if chr(x.id1.comp) in ("a", "c")] # don't need delete
        if not tors: continue
        gr = gemmi.ChemComp.read_group(mod.group_id)
        if mod.comp_id and mod.comp_id in all_tors["mon"]:
            all_tors["mon"][mod.comp_id].extend(tors)
        elif not mod.comp_id and gr in groups:
            for mon_id in groups[gr]:
                all_tors["mon"][mon_id].extend(tors)
    for lnk_id in monlib.links:
        lnk = monlib.links[lnk_id]
        if lnk.rt.torsions:
            all_tors["link"][lnk_id] = [x.label for x in lnk.rt.torsions]
    for k in all_tors:
        for kk in all_tors[k]:
            all_tors[k][kk] = set(all_tors[k][kk])
            
    # Apply include/exclude rule
    for mon in all_tors["mon"]:
        match_f = lambda r: ("tors_name" in r and
                             ("residue" in r and fnmatch.fnmatch(mon, r["residue"]) or
                              mon in groups.get(gemmi.ChemComp.read_group(r.get("group", "")), [])))
        use_tors = []
        for r in include_rules:
            if match_f(r):
                use_tors.extend(x for x in all_tors["mon"][mon] if fnmatch.fnmatch(x, r["tors_name"]))
        for r in exclude_rules:
            if match_f(r):
                use_tors = [x for x in use_tors if not fnmatch.fnmatch(x, r["tors_name"])]
        if use_tors:
            ret["monomer"][mon] = sorted(use_tors)
    for lnk in all_tors["link"]:
        match_f = lambda r: ("tors_name" in r and
                             "link" in r and fnmatch.fnmatch(lnk, r["link"]))
        use_tors = []
        for r in include_rules:
            if match_f(r):
                use_tors.extend(x for x in all_tors["link"][lnk] if fnmatch.fnmatch(x, r["tors_name"]))
        for r in exclude_rules:
            if match_f(r):
                use_tors = [x for x in use_tors if not fnmatch.fnmatch(x, r["tors_name"])]
        if use_tors:
            ret["link"][lnk] = sorted(use_tors)

    return ret
# select_restrained_torsions()

def prepare_topology(st, monlib, h_change, ignore_unknown_links=False, raise_error=True, check_hydrogen=False,
                     use_cispeps=False, add_metal_restraints=True, params=None):
    # Check duplicated atoms
    bad = []
    for chain in st[0]:
        bad_res = []
        for res  in chain:
            n_uniq = len({(a.name, a.altloc) for a in res})
            if n_uniq != len(res):
                bad_res.append(str(res.seqid))
        if bad_res:
            bad.append(" chain {}: {}".format(chain.name, " ".join(bad_res)))
    if bad:
        raise RuntimeError("Following residues have duplicated atoms. Check your model.\n{}".format("\n".join(bad)))

    if add_metal_restraints:
        metalc = MetalCoordination(monlib)
        keywords, todel = metalc.setup_restraints(st)
        con_bak = []
        for i in sorted(todel, reverse=True):
            # temporarily remove connection not to put a bond restraint
            con = st.connections.pop(i)
            con_bak.append((i, con))
            # flag non-hydrogen
            cra2 = st[0].find_cra(con.partner2, ignore_segment=True)
            cra2.atom.calc_flag = gemmi.CalcFlag.NoHydrogen
        if params:
            parsed = refmac_keywords.parse_keywords(keywords).get("exte")
            if parsed:
                params["exte"] = params.get("exte", []) + parsed
    else:
        keywords = []
    # these checks can be done after sorting links
    logger.writeln("Creating restraints..")
    sio = io.StringIO()
    topo = gemmi.prepare_topology(st, monlib, h_change=h_change, warnings=sio, reorder=False,
                                  ignore_unknown_links=ignore_unknown_links, use_cispeps=use_cispeps)
    for l in sio.getvalue().splitlines(): logger.writeln(" " + l)
    unknown_cc = set()
    link_related = set()
    nan_hydr = set()

    def extra_defined(res1, res2): # TODO should check alt
        for link in topo.extras:
            res12 = (link.res1, link.res2)
            if link.link_id and (res12 == (res1, res2) or res12 == (res2, res1)):
                return True
        return False

    # collect info
    info = {}
    for cinfo in topo.chain_infos:
        toadd = info.setdefault(cinfo.chain_ref.name, {})
        if cinfo.polymer:
            gaps = []
            for rinfo in cinfo.res_infos:
                if (rinfo.prev and rinfo.prev[0].link_id in ("gap", "") and
                    not extra_defined(rinfo.prev[0].res1, rinfo.prev[0].res2)):
                    gaps.append((rinfo.prev[0].res1, rinfo.prev[0].res2))
            toadd["polymer"] = (str(cinfo.polymer_type).replace("PolymerType.", ""),
                                cinfo.res_infos[0].res.seqid,
                                cinfo.res_infos[-1].res.seqid,
                                len(cinfo.res_infos), gaps)
        else:
            l = toadd.setdefault("nonpolymer", [])
            for ri in cinfo.res_infos:
                l.append(ri.res.name)
    logger.writeln("\nChain info:")
    for chain in info:
        logger.writeln(" chain {}".format(chain))
        if "polymer" in info[chain]:
            logger.writeln("  {}: {}..{} ({} residues)".format(*info[chain]["polymer"][:-1]))
            for gap in info[chain]["polymer"][-1]:
                logger.writeln("    gap between {} and {}".format(*gap))
        if "nonpolymer" in info[chain]:
            n_res = len(info[chain]["nonpolymer"])
            uniq = set(info[chain]["nonpolymer"])
            logger.writeln("  ligands: {} ({} residues)".format(" ".join(uniq), n_res))
    logger.writeln("")
    
    for cinfo in topo.chain_infos:
        for rinfo in cinfo.res_infos:
            cc_org = monlib.monomers[rinfo.res.name] if rinfo.res.name in monlib.monomers else None
            for ia in reversed(range(len(rinfo.res))):
                atom = rinfo.res[ia]
                atom_str = "{}/{} {}/{}".format(cinfo.chain_ref.name, rinfo.res.name, rinfo.res.seqid, atom.name)
                cc = rinfo.get_final_chemcomp(atom.altloc)
                if not cc.find_atom(atom.name):
                    # warning message should have already been given by gemmi
                    if cc_org and cc_org.find_atom(atom.name):
                        if check_hydrogen or not atom.is_hydrogen():
                            link_related.add(rinfo.res.name)
                    else:
                        if check_hydrogen or not atom.is_hydrogen():
                            unknown_cc.add(rinfo.res.name)
                
                if atom.is_hydrogen() and atom.calc_flag == gemmi.CalcFlag.Dummy:
                    logger.writeln(" Warning: hydrogen {} could not be added - Check dictionary".format(atom_str))
                    unknown_cc.add(rinfo.res.name)
                elif any(numpy.isnan(atom.pos.tolist())): # TODO add NaN test before prepare_toplogy
                    logger.writeln(" Warning: {} position NaN!".format(atom_str))
                    nan_hydr.add(rinfo.res.name)

    if raise_error and (unknown_cc or link_related):
        msgs = []
        if unknown_cc: msgs.append("restraint cif file(s) for {}".format(",".join(unknown_cc)))
        if link_related: msgs.append("proper link cif file(s) for {} or check your model".format(",".join(link_related)))
        raise RuntimeError("Provide {}".format(" and ".join(msgs)))
    if raise_error and nan_hydr:
        raise RuntimeError("Some hydrogen positions became NaN. The geometry of your model may be of low quality. Consider not adding hydrogen")
    if not use_cispeps:
        topo.set_cispeps_in_structure(st)
    if add_metal_restraints:
        for i, con in sorted(con_bak):
            st.connections.insert(i, con)
    return topo, keywords
# prepare_topology()

def check_monlib_support_nucleus_distances(monlib, resnames):
    good = True
    nucl_not_found = []
    for resn in resnames:
        if resn not in monlib.monomers:
            logger.error("ERROR: monomer information of {} not loaded".format(resn))
            good = False
        else:
            mon = monlib.monomers[resn]
            no_nuc = False
            for bond in mon.rt.bonds:
                is_h = (mon.get_atom(bond.id1.atom).is_hydrogen(), mon.get_atom(bond.id2.atom).is_hydrogen())
                if any(is_h) and bond.value_nucleus != bond.value_nucleus:
                    no_nuc = True
                    break
            if no_nuc:
                nucl_not_found.append(resn)
                good = False

    if nucl_not_found:
        logger.writeln("WARNING: nucleus distance is not found for: {}".format(" ".join(nucl_not_found)))
        logger.writeln("         default scale ({}) is used for nucleus distances.".format(default_proton_scale))
    return good
# check_monlib_support_nucleus_distances()

def remove_duplicated_links(connections):
    # ignore p.res_id.name?
    totuple = lambda p: (p.chain_name, p.res_id.seqid.num, p.res_id.seqid.icode, p.atom_name, p.altloc)
    dic = {}
    for i, con in enumerate(connections):
        dic.setdefault(tuple(sorted([totuple(con.partner1), totuple(con.partner2)])), []).append(i)
    todel = []
    for k in dic:
        if len(dic[k]) > 1:
            ids = set(connections[c].link_id for c in dic[k] if connections[c].link_id.strip())
            if len(ids) > 1:
                logger.writeln(" WARNING: duplicated links are found with different link_id")
            tokeep = dic[k][0]
            if ids:
                for c in dic[k]:
                    if connections[c].link_id.strip():
                        tokeep = c
                        break
            todel.extend(c for c in dic[k] if c != tokeep)

    for i in sorted(todel, reverse=True):
        del connections[i]
    if todel:
        logger.writeln(" {} duplicated links were removed.".format(len(todel)))
# remove_duplicated_links()

def find_and_fix_links(st, monlib, bond_margin=1.3, find_metal_links=True, add_found=True, find_symmetry_related=True,
                       metal_margin=1.1, add_only_from=None):
    metalc = MetalCoordination(monlib)
    """
    Identify link ids for st.connections and find new links
    This is required for correctly recognizing link in gemmi.prepare_topology
    Note that it ignores segment IDs
    FIXME it assumes only one bond exists in a link. It may not be the case in future.
    """
    from servalcat.utils import model

    logger.writeln("Checking links defined in the model")
    remove_duplicated_links(st.connections)
    for con in st.connections:
        if con.type == gemmi.ConnectionType.Hydrog: continue
        if con.link_id == "gap": continue # TODO check residues?
        cra1, cra2 = st[0].find_cra(con.partner1, ignore_segment=True), st[0].find_cra(con.partner2, ignore_segment=True)
        if None in (cra1.atom, cra2.atom):
            logger.writeln(" WARNING: atom(s) not found for link: id= {} atom1= {} atom2= {}".format(con.link_id, con.partner1, con.partner2))
            continue
        if cra1.atom.element.is_metal or cra2.atom.element.is_metal:
            con.type = gemmi.ConnectionType.MetalC
        if con.asu != gemmi.Asu.Same: # XXX info from metadata may be wrong
            nimage = st.cell.find_nearest_image(cra1.atom.pos, cra2.atom.pos, con.asu)
            image_idx = nimage.sym_idx
            dist = nimage.dist()
        else:
            image_idx = 0
            con.asu = gemmi.Asu.Same
            dist = cra1.atom.pos.dist(cra2.atom.pos)
        con.reported_distance = dist
        atoms_str = "atom1= {} atom2= {} image= {}".format(cra1, cra2, image_idx)
        if con.link_id:
            link = monlib.get_link(con.link_id)
            inv = False
            if link is None:
                logger.writeln(" WARNING: link {} not found in the library. Please provide link dictionary.".format(con.link_id))
                con.link_id = "" # let gemmi find proper link in prepare_topology()
                continue
            else:
                match, _, _ = monlib.test_link(link, cra1.residue.name, cra1.atom.name, cra2.residue.name, cra2.atom.name)
                if not match and monlib.test_link(link, cra2.residue.name, cra2.atom.name, cra1.residue.name, cra1.atom.name)[0]:
                    match = True
                    inv = True
                if not match:
                    logger.writeln(" WARNING: link id and atoms mismatch: id= {} {}".format(link.id, atoms_str))
                    continue
        else:
            link, inv, _, _ = monlib.match_link(cra1.residue, cra1.atom.name, cra1.atom.altloc,
                                                cra2.residue, cra2.atom.name, cra2.atom.altloc)
            if link:
                con.link_id = link.id
            elif find_metal_links and con.type == gemmi.ConnectionType.MetalC:
                logger.writeln(" Metal link will be added: {} dist= {:.2f}".format(atoms_str, dist))
                if cra2.atom.element.is_metal:
                    inv = True # make metal first
            else:
                ideal_dist = monlib.find_ideal_distance(cra1, cra2)
                logger.writeln(" Link unidentified (simple bond will be used): {} dist= {:.2f} ideal= {:.2f}".format(atoms_str,
                                                                                                                     dist,
                                                                                                                     ideal_dist))
                continue
        if link:
            logger.writeln(" Link confirmed: id= {} {} dist= {:.2f} ideal= {:.2f}".format(link.id,
                                                                                          atoms_str,
                                                                                          dist,
                                                                                          link.rt.bonds[0].value))
            if con.link_id == "disulf":
                con.type = gemmi.ConnectionType.Disulf
        if inv:
            con.partner1 = model.cra_to_atomaddress(cra2)
            con.partner2 = model.cra_to_atomaddress(cra1)
    if len(st.connections) == 0:
        logger.writeln(" no links defined in the model")

    logger.writeln("Finding new links (will be added if marked by *)")
    ns = gemmi.NeighborSearch(st[0], st.cell, 5.).populate()
    cs = gemmi.ContactSearch(4.)
    cs.ignore = gemmi.ContactSearch.Ignore.AdjacentResidues # may miss polymer links not contiguous in a chain?
    results = cs.find_contacts(ns)
    onsb = set(gemmi.Element(x) for x in "ONSB")
    n_found = 0
    for r in results:
        if st.find_connection_by_cra(r.partner1, r.partner2, ignore_segment=True): continue
        link, inv, _, _ = monlib.match_link(r.partner1.residue, r.partner1.atom.name, r.partner1.atom.altloc,
                                            r.partner2.residue, r.partner2.atom.name, r.partner2.atom.altloc,
                                            (r.dist / 1.4)**2)
        if link is None and r.partner2.atom.element.is_metal:
            inv = True # make metal first
        if inv:
            cra1, cra2 = r.partner2, r.partner1
        else:
            cra1, cra2 = r.partner1, r.partner2
        im = st.cell.find_nearest_pbc_image(cra1.atom.pos, cra2.atom.pos, r.image_idx)
        #assert r.image_idx == im.sym_idx # should we check this?
        if not find_symmetry_related and not im.same_asu():
            continue
        atoms_str = "atom1= {} atom2= {} image= {}".format(cra1, cra2, r.image_idx)
        if im.pbc_shift != (0,0,0):
            atoms_str += " ({},{},{})".format(*im.pbc_shift)
        if link:
            if r.dist > link.rt.bonds[0].value * bond_margin: continue
            will_be_added = add_found and (not add_only_from or link.id in add_only_from)
            logger.writeln(" {}New link found: id= {} {} dist= {:.2f} ideal= {:.2f}".format("*" if will_be_added else " ",
                                                                                            link.id,
                                                                                            atoms_str,
                                                                                            r.dist,
                                                                                            link.rt.bonds[0].value))
        elif find_metal_links:
            # link only metal - O/N/S/B
            if r.partner1.atom.element.is_metal == r.partner2.atom.element.is_metal: continue
            if not cra2.atom.element in onsb: continue
            max_ideal = metalc.find_max_dist(cra1, cra2)
            if r.dist > max_ideal * metal_margin: continue # tolerance should be smaller than that for other links
            will_be_added = add_found
            logger.writeln(" {}Metal link found: {} dist= {:.2f} max_ideal= {:.2f}".format("*" if will_be_added else " ",
                                                                                           atoms_str,
                                                                                           r.dist, max_ideal))
        n_found += 1
        if not will_be_added: continue
        con = gemmi.Connection()
        con.name = "added{}".format(n_found)
        if link:
            con.link_id = link.id
            con.type = gemmi.ConnectionType.Disulf if link.id == "disulf" else gemmi.ConnectionType.Covale
        if cra1.atom.element.is_metal or cra2.atom.element.is_metal:
            con.type = gemmi.ConnectionType.MetalC
        con.asu = gemmi.Asu.Same if im.same_asu() else gemmi.Asu.Different
        con.partner1 = model.cra_to_atomaddress(cra1)
        con.partner2 = model.cra_to_atomaddress(cra2)
        con.reported_distance = r.dist
        st.connections.append(con)
    if n_found == 0:
        logger.writeln(" no links found")
# find_and_fix_links()

def add_hydrogens(st, monlib, pos="elec"):
    assert pos in ("elec", "nucl")
    topo = prepare_topology(st, monlib, h_change=gemmi.HydrogenChange.ReAddButWater, ignore_unknown_links=True)
    
    if pos == "nucl":
        logger.writeln("Generating hydrogens at nucleus positions")
        resnames = st[0].get_all_residue_names()
        check_monlib_support_nucleus_distances(monlib, resnames)
        topo.adjust_hydrogen_distances(gemmi.Restraints.DistanceOf.Nucleus, default_scale=default_proton_scale)
    else:
        logger.writeln("Generating hydrogens at electron positions")
# add_hydrogens()

def make_atom_spec(cra):
    chain = cra.chain.name
    resi = cra.residue.seqid.num
    ins = cra.residue.seqid.icode
    atom = cra.atom.name
    s = "chain {} resi {} ins {} atom {}".format(chain, resi, ins if ins.strip() else ".", atom)
    if cra.atom.altloc != "\0":
        s += " alt {}".format(cra.atom.altloc)
    return s
# make_atom_spec()        

def dictionary_block_names(monlib, topo):
    used = {x.lower() for x in monlib.monomers}
    for chain_info in topo.chain_infos:
        for res_info in chain_info.res_infos:
            for link in res_info.prev:
                # won't be included if the name starts with "auto-", but don't do such checks here
                used.add("link_" + link.link_id.lower())
        for mod in res_info.mods:
            used.add("mod_" + mod.id.lower())
    for extra in topo.extras:
        used.add("link_" + extra.link_id.lower())
    return used
# dictionary_block_names()

def prepare_ncs_restraints(st, rms_loc_nlen=5, min_nalign=10, max_rms_loc=2.0):
    logger.writeln("Finding NCS..")
    polymers = {}
    for chain in st[0]:
        rs = chain.get_polymer()
        p_type = rs.check_polymer_type()
        if p_type in (gemmi.PolymerType.PeptideL, gemmi.PolymerType.PeptideD,
                      gemmi.PolymerType.Dna, gemmi.PolymerType.Rna, gemmi.PolymerType.DnaRnaHybrid):
            polymers.setdefault(p_type, []).append((chain, rs))

    scoring = gemmi.AlignmentScoring()
    scoring.match = 0
    scoring.mismatch = -1
    scoring.gapo = 0
    scoring.gape = -1
            
    al_res = []
    ncslist = ext.NcsList()
    for pt in polymers:
        #print(pt, [x[0].name for x in polymers[pt]])
        pols = polymers[pt]
        for i in range(len(pols)-1):
            q = [x.name for x in pols[i][1]]
            for j in range(i+1, len(pols)):
                al = gemmi.align_sequence_to_polymer(q, pols[j][1], pt, scoring)
                if al.match_count < min_nalign: continue
                su = gemmi.calculate_superposition(pols[i][1], pols[j][1], pt, gemmi.SupSelect.All)
                obj = ext.NcsList.Ncs(al, pols[i][1], pols[j][1])
                obj.calculate_local_rms(rms_loc_nlen)
                if len(obj.local_rms) == 0: continue
                ave_local_rms = numpy.mean(obj.local_rms)
                if ave_local_rms > max_rms_loc: continue
                ncslist.ncss.append(obj)
                al_res.append({"chain_1": "{} ({}..{})".format(pols[i][0].name, pols[i][1][0].seqid, pols[i][1][-1].seqid),
                               "chain_2": "{} ({}..{})".format(pols[j][0].name, pols[j][1][0].seqid, pols[j][1][-1].seqid),
                               "aligned": al.match_count,
                               "identity": al.calculate_identity(1),
                               "rms": su.rmsd,
                               "ave(rmsloc)": ave_local_rms,
                               })
    ncslist.set_pairs()
    df = pandas.DataFrame(al_res)
    df.index += 1
    logger.writeln(df.to_string(float_format="%.2f"))
    return ncslist
# prepare_ncs_restraints()

class MetalCoordination:
    def __init__(self, monlib, dbfile=None):
        self.monlib = monlib
        if dbfile is None:
            dbfile = os.path.join(monlib.path(), "metals.json")
        if os.path.exists(dbfile):
            self.metals = json.load(open(dbfile))["metal_coordination"]
        else:
            self.metals = {}
            logger.writeln("WARNING: {} not found".format(dbfile))
    # __init__()

    def find_max_dist(self, cra_metal, cra_ligand):
        vals = self.find_ideal_distances(cra_metal.atom.element, cra_ligand.atom.element)
        if len(vals) == 0:
            # if not found
            return self.monlib.find_ideal_distance(cra_metal, cra_ligand)
        return max(x["median"] for x in vals)
    # find_max_dist()

    def find_ideal_distances(self, el_metal, el_ligand):
        ideals = {}
        if el_metal.name not in self.metals or el_ligand.name not in self.metals[el_metal.name]:
            return []
        return self.metals[el_metal.name][el_ligand.name]
    # find_ideal_distances
    
    def setup_restraints(self, st):
        ret = [] # returns Refmac keywords
        lookup = {x.atom: x for x in st[0].all()}
        coords = {}
        todel = []
        for i, con in enumerate(st.connections):
            if con.link_id == "" and con.type == gemmi.ConnectionType.MetalC:
                cra1 = st[0].find_cra(con.partner1, ignore_segment=True)
                cra2 = st[0].find_cra(con.partner2, ignore_segment=True)
                if None in (cra1.atom, cra2.atom): continue
                ener_ideal = self.monlib.find_ideal_distance(cra1, cra2)
                coords.setdefault(cra1.atom.element, {}).setdefault(cra1.atom, []).append((cra2.atom, i, ener_ideal))
        if coords:
            logger.writeln("Metal coordinations detected")
        for metal in coords:
            logger.writeln(" Metal: {}".format(metal.name))
            ligand_els = {x[0].element for m in coords[metal] for x in coords[metal][m]}
            logger.writeln("  ideal distances")
            ideals = {}
            for el in ligand_els:
                logger.write("   {}: ".format(el.name))
                vals = self.find_ideal_distances(metal, el)
                if len(vals) == 0:
                    ener_ideals = {x[2] for m in coords[metal] for x in coords[metal][m] if x[0].element == el}
                    logger.write(" ".join("{:.2f}".format(x) for x in ener_ideals))
                    logger.writeln(" (from ener_lib)")
                else:
                    logger.writeln(" ".join("{:.4f} ({} coord)".format(x["median"], x["coord"]) for x in vals))
                    ideals[el] = [(x["median"], x["mad"]) for x in vals if x["mad"] > 0]
            logger.writeln("")
            for i, am in enumerate(coords[metal]):
                logger.writeln("  site {}: {}".format(i+1, lookup[am]))
                for j, (lig, con_idx, _) in enumerate(coords[metal][am]):
                    con = st.connections[con_idx]
                    logger.writeln("    ligand {}: {} dist= {:.2f}".format(j+1, lookup[lig],
                                                                           con.reported_distance))
                    specs = [make_atom_spec(x) for x in (lookup[am], lookup[lig])]
                    if lig.element not in ideals:
                        continue
                    todel.append(con_idx)
                    for k, (ideal, sigma) in enumerate(ideals[lig.element]):
                        exte_str  = "exte dist first {} seco {} ".format(*specs)
                        exte_str += "valu {:.4f} sigm {:.4f} type 1 ".format(ideal, sigma)
                        if con.asu == gemmi.Asu.Different:
                            exte_str += "symm y"
                        ret.append(exte_str)
                        #b = ext.Geometry.Bond(am, lig)
                        #b.values.append(ext.Geometry.Bond.Value(ideal, sigma, ideal, sigma))
                        #b.type = 0 if k == 0 else 1
                        #ret.append(b)
                logger.writeln("")
        return ret, list(set(todel))
    # setup_restraints()



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import gemmi
import subprocess
import numpy
import re
from servalcat.utils import logger
from servalcat.utils import fileio
from servalcat.utils import model
from servalcat.utils import generate_operators

def add_symmetry_args(parser, require_pg=False):
    parser.add_argument('--pg', required=require_pg, help="Point group symbol")
    parser.add_argument('--twist', type=float, help="Helical twist (degree)")
    parser.add_argument('--rise', type=float, help="Helical rise (Angstrom)")
    parser.add_argument('--center', type=float, nargs=3, help="Origin of symmetry. Default: center of the box")
    parser.add_argument('--axis1', type=float, nargs=3, help="Axis1 (if I: 5-fold, O: 4-fold, T: 3-fold)")
    parser.add_argument('--axis2', type=float, nargs=3, help="Axis2 (if I: 5-fold, O: 4-fold, T: 3-fold, Dn: 2-fold)")
# add_symmetry_args()
    
def update_ncs_from_args(args, st, map_and_start=None, filter_contacting=False,
                         helical_min_n=None, helical_max_n=None):
    is_helical = args.twist is not None
    if not is_helical and not args.pg:
        if len(st.ncs) > 0:
            logger.writeln("Strict NCS detected from model.")
            show_ncs_operators_axis_angle(st.ncs)
        return
    
    if len(st.ncs) > 0:
        logger.writeln(" WARNING: NCS information in model file will be ignored")

    ncsops = ncsops_from_args(args, st.cell, map_and_start=map_and_start, st=st,
                              helical_min_n=helical_min_n, helical_max_n=helical_max_n)

    st.ncs.clear()
    st.ncs.extend([x for x in ncsops if not x.tr.is_identity()])

    if filter_contacting:
        model.filter_contacting_ncs(st)
# update_ncs_from_args()

def ncsops_from_args(args, cell, map_and_start=None, st=None, helical_min_n=None, helical_max_n=None):
    is_helical = args.twist is not None
    
    if map_and_start is not None:
        start_xyz = numpy.array(map_and_start[0].get_position(*map_and_start[1]).tolist())
    else:
        start_xyz = numpy.zeros(3)

    if args.center is None:
        A = numpy.array(cell.orthogonalization_matrix.tolist())
        center = numpy.sum(A, axis=1) / 2 #+ start_xyz
        logger.writeln("Center: {}".format(center))
    else:
        center = numpy.array(args.center)
        
    if is_helical:
        ncsops = generate_helical_operators(start_xyz, center,
                                            args.pg, args.twist, args.rise,
                                            axis1=args.axis1, axis2=args.axis2,
                                            st=st, min_n=helical_min_n, max_n=helical_max_n)
        logger.writeln("{} helical operators found".format(len(ncsops)))
    else:
        _, _, ops = operators_from_symbol(args.pg, axis1=args.axis1, axis2=args.axis2)
        logger.writeln("{} operators found for {}".format(len(ops), args.pg))
        show_operators_axis_angle(ops)
        ncsops = make_NcsOps_from_matrices(ops, cell=cell, center=center)

    return ncsops
# ncsops_from_args()

def get_matrices_using_relion(sym):
    ps = subprocess.check_output(["relion_refine", "--sym", sym.strip(), "--print_symmetry_ops"])

    ret = []
    read_flag = -1
    for l in ps.splitlines():
        if b"R(" in l:
            ret.append(numpy.zeros((3,3)))
            read_flag = 0
        elif 0 <= read_flag < 3:
            ret[-1][read_flag,:] = [float(x) for x in l.split()]
            read_flag += 1
        elif read_flag >= 3:
            read_flag = -1

    return ret
# get_matrices_using_relion()

def operators_from_symbol(op, axis1=None, axis2=None):
    r = re.search("^([CDITO])([0-9]*)$", op.upper())
    if not r:
        raise RuntimeError("Invalid point group symbol: {}".format(op))
    a, n = r.groups()
    if n:
        n = int(n)
        if n <= 0:
            raise RuntimeError("Non positive order given: {}".format(op))
        elif a in ("T", "O"):
            raise RuntimeError("You cannot give number after T,O: {}".format(op))
        elif a == "I" and n > 4:
            raise RuntimeError("Only I1-4 are supported for I variants: {}".format(op))

    if axis1 is not None: axis1 = numpy.array(axis1)
    if axis2 is not None: axis2 = numpy.array(axis2)

    # RELION's conventions
    if a == "I":
        order1 = order2 = 5
        if not n or n == 2:
            if axis1 is None: axis1 = numpy.array([0, 0.85065, 0.52573])
            if axis2 is None: axis2 = numpy.array([0.52573, 0, 0.85065])
        elif n == 1:
            if axis1 is None: axis1 = numpy.array([0.85065, 0., 0.52573])
            if axis2 is None: axis2 = numpy.array([0.52573, 0.85065, 0.])
        elif n == 3:
            if axis1 is None: axis1 = numpy.array([0.723606, 0.525732, 0.447214])
            if axis2 is None: axis2 = numpy.array([-0.276391, -0.850645, 0.447216])
        elif n == 4:
            if axis1 is None: axis1 = numpy.array([-0.723606, -0.525732, 0.447214])
            if axis2 is None: axis2 = numpy.array([0., 0., 1.])
    elif a == "O":
        order1 = order2 = 4
        if axis1 is None: axis1 = numpy.array([0,0,1.0])
        if axis2 is None: axis2 = numpy.array([0,1.0,0])
    elif a == "T":
        order1 = order2 = 3
        if axis1 is None: axis1 = numpy.array([0.0,0.0,1.0])
        if axis2 is None: axis2 = numpy.array([0.0,0.94280904,-0.33333333])
    elif a == "D":
        order1 = n
        if axis1 is None: axis1 = numpy.array([0,0,1.0])
        order2 = 2
        if axis2 is None: axis2 = numpy.array([1.,0.,0.])
    elif a == "C":
        order1 = n
        if axis1 is None: axis1 = numpy.array([0,0,1.0])
        order2 = 0
        axis2 = None
    return generate_operators.generate_all_elements(axis1, order1, axis2, order2)
# operators_from_symbol()

def show_operators_axis_angle(ops):
    for i, op in enumerate(ops):
        ax, ang = generate_operators.Rotation2AxisAngle_general(op)
        logger.writeln(" operator {:3d} angle= {:7.3f} deg axis= {}".format(i+1, numpy.rad2deg(ang), list(ax)))
# show_operators_axis_angle()

def show_ncs_operators_axis_angle(ops):
    # ops: List of gemmi.NcsOp
    for i, op in enumerate(ops):
        op2 = numpy.array(op.tr.mat.tolist())
        ax, ang = generate_operators.Rotation2AxisAngle_general(op2)
        axlab = "[{: .4f}, {: .4f}, {: .4f}]".format(*ax)
        trlab = "[{: 9.4f}, {: 9.4f}, {: 9.4f}]".format(*op.tr.vec.tolist())
        logger.writeln(" operator {:3s} angle= {:7.3f} deg axis= {} trans= {} {}".format(op.id, numpy.rad2deg(ang),
                                                                                       axlab, trlab,
                                                                                       "given" if op.given else ""))
# show_operators_axis_angle()

def read_helical_parameters_from_mmcif(cif_in):
    doc = fileio.read_cif_safe(cif_in)
    b = list(filter(lambda b: b.find_loop("_atom_site.id"), doc))[0]
    deltaphi = b.find_value("_em_helical_entity.angular_rotation_per_subunit")
    deltaz = b.find_value("_em_helical_entity.axial_rise_per_subunit")
    deltaphi, deltaz = float(deltaphi), float(deltaz)
    axsym = b.find_value("_em_helical_entity.axial_symmetry")
    return axsym, deltaphi, deltaz
# read_helical_parameters_from_mmcif()

def generate_helical_operators(start_xyz, center, axsym, deltaphi, deltaz, axis1=None, axis2=None,
                               st=None, min_n=None, max_n=None, padding=2.):
    if not axsym: axsym = "C1"
    if axis1 is None: axis1 = numpy.array([0,0,1.])
    else: axis1 /= numpy.linalg.norm(axis1)
    _, _, axtrs = operators_from_symbol(axsym, axis1, axis2)
    if min_n is None or max_n is None:
        assert st is not None
        all_z = numpy.dot([cra.atom.pos.tolist() for cra in st[0].all()], axis1)
        min_z, max_z = numpy.min(all_z), numpy.max(all_z)
        direc = numpy.argmax([numpy.dot(axis1, v) for v in ((1.,0,0), (0,1.,0), (0,0,1.))]) # assume axis1 along any of a,b,c axis
        if min_n is None:
            min_n = -int((min_z - padding - start_xyz[direc]) / deltaz)
        if max_n is None:
            max_n = int((st.cell.parameters[direc] + start_xyz[direc] - max_z - padding) / deltaz)
    if min_n > max_n:
        min_n, max_n = max_n, min_n
    ops = []
    for i in range(min_n, max_n+1):
        deg = deltaphi*i
        t = numpy.deg2rad(deg)
        m = generate_operators.AngleAxis2rotatin(axis1, t)
        s = numpy.array(axis1 * deltaz*i)
        for a in axtrs:
            mat = numpy.dot(m, a)
            news = s + numpy.dot(mat, -center) + center
            tr = gemmi.Transform(gemmi.Mat33(mat), gemmi.Vec3(*news))
            newop = gemmi.NcsOp(tr, str(len(ops)+1), tr.is_identity())
            ops.append(newop)

    return ops
# generate_helical_operators()

def make_NcsOps_from_matrices(matrices, cell=None, center=None):
    if center is None:
        A = numpy.array(cell.orthogonalization_matrix.tolist())
        center = numpy.sum(A,axis=1) / 2

    center = gemmi.Vec3(*center)
    ops = []
    for i, m in enumerate(matrices):
        m = gemmi.Mat33(m)
        transl = m.multiply(-center) + center
        op = gemmi.NcsOp(gemmi.Transform(m, transl), str(i+1), m.is_identity())
        ops.append(op)
        
    return ops
# make_NcsOps_from_matrices()

def find_center_of_origin(mat, vec): # may not be unique.
    tmp = numpy.identity(3) - numpy.array(mat)
    ret = numpy.dot(numpy.linalg.pinv(tmp), vec.tolist())
    resid = vec.tolist() - (numpy.dot(mat, -ret) + ret)
    return gemmi.Vec3(*ret), gemmi.Vec3(*resid)
# find_center_of_origin()

def ncs_ops_for_refmac(ncs_ops):
    def make_line(tr):
        m = tr.mat.tolist()
        m_str = " ".join([str(m[i][j]) for i in range(3) for j in range(3)])
        t_str = " ".join([str(x) for x in tr.vec.tolist()])
        return "ncsc matrix {} {}".format(m_str, t_str)
    
    ret = []
    # REFMAC requires identity op
    if not any(x.tr.is_identity() for x in ncs_ops):
        ret.append(make_line(gemmi.Transform()))

    for op in ncs_ops:
        if not op.given: ret.append(make_line(op.tr))
    return ret
# ncs_ops_for_refmac()

# TODO def euler2matrix(euler):
# TODO def polar2matrix(polar):

def parse_ncsc_keywords(kwd_str):
    # FIXME handle lines ending with -
    lines = kwd_str.splitlines()
    ret = []
    for l in lines:
        l = l.split()
        if len(l) < 3: continue
        if l[0].lower().startswith("ncsc"):
            if l[1].lower().startswith("matr"):
                vals = [float(x) for x in l[2:]]
                if len(vals) != 12:
                    print("Bad nsc matrix line: {}".format(" ".join(l)))
                    continue
                op = gemmi.NcsOp()
                op.tr.mat.fromlist([vals[3*x:3*x+3] for x in range(3)])
                op.tr.vec.fromlist(vals[9:])
                op.id = str(len(ret)+1)
                op.given = op.tr.is_identity()
                ret.append(op)
            elif l[1].lower().startswith("eule"):
                pass # TODO
            elif l[1].lower().startswith("pola"):
                pass # TODO
    return ret
# parse_ncsc_keywords()

def apply_shift_for_ncsops(ncsops, shift):
    new_ops = []
    s = gemmi.Vec3(*shift)
    for op in ncsops:
        newt = op.tr.vec + s - op.tr.mat.multiply(s)
        newop = gemmi.NcsOp(gemmi.Transform(op.tr.mat, newt), op.id, op.given)
        new_ops.append(newop)

    return new_ops
# apply_shift_for_ncsops()

def write_symmetry_expanded_model(st, prefix, pdb=False, cif=False, cif_ref=None):
    st_new = st.clone()
    model.expand_ncs(st_new)
    fileio.write_model(st_new, prefix=prefix, pdb=pdb, cif=cif, cif_ref=cif_ref)
# write_symmetry_expanded_model()
